/*
   Class Name:

      COctree.

   Created by:

      Allen Sherrod (Programming Ace of www.UltimateGameProgramming.com).

   Description:

      This class is used to store a mesh in a octree.
*/


#include<string.h>
#include"COctree.h"


int g_currentLevel = 0;
int g_maxLevel = 0;
int g_maxTriangles = 0;


COctree::COctree() : m_isLeaf(0), m_vertices(0), m_numPoints(0), m_texCoord1(0),
                     m_texCoord2(0), m_bbSize(0)
{
   m_bbCenter[0] = m_bbCenter[1] = m_bbCenter[2] = 0;
   for(int i = 0; i < 8; i++) m_subNodes[i] = NULL;
}


COctree::~COctree()
{
   Shutdown();
}


bool COctree::Initialize(float *vertexList, int numVerts, float *t1, float *t2,
                         int maxLevels, int maxTris)
{
   g_maxLevel = maxLevels;
   g_maxTriangles = maxTris;
   m_numPoints = numVerts;

   // Calculate the bounding box then recursively create
   // all child nodes.
   CalculateBoundingBox(vertexList, numVerts);
   CreateNode(vertexList, numVerts, t1, t2, m_bbCenter, m_bbSize);
   return true;
}


void COctree::CalculateBoundingBox(float *vertexList, int numVerts)
{
   float maxWidth = 0, maxHeight = 0, maxDepth = 0;

   if(vertexList == NULL || numVerts <= 0) return;

   m_bbCenter[0] = m_bbCenter[1] = m_bbCenter[2] = 0;

   // Loop through and add up all axises.
   for(int i = 0; i < numVerts * 3; i+=3)
      {
         m_bbCenter[0] = m_bbCenter[0] + vertexList[i];
         m_bbCenter[1] = m_bbCenter[1] + vertexList[i + 1];
         m_bbCenter[2] = m_bbCenter[2] + vertexList[i + 2];
      }

   // Divide by the number of points to get the center.
   m_bbCenter[0] /= numVerts;
   m_bbCenter[1] /= numVerts;
   m_bbCenter[2] /= numVerts;

   // Loop through and get the axis which is farthest from
   // the center of the node.
   for(i = 0; i < numVerts * 3; i+=3)
      {
         // Get the distance of this vertex to the center.
         float width = vertexList[i + 0] - m_bbCenter[0];
         float height = vertexList[i + 1] - m_bbCenter[1];
         float depth = vertexList[i + 2] - m_bbCenter[2];

         // Make sure they are not negative numbers.
         if(width < 0) width *= -1;
         if(height < 0) height *= -1;
         if(depth < 0) depth *= -1;

         // Update if any of these are bigger than the recorded max.
         if(width > maxWidth) maxWidth = width;
         if(height > maxHeight) maxHeight = height;
         if(depth > maxDepth) maxDepth = depth;
      }

   // Convert from radius distance to diameter of each axis.
   maxWidth *= 2;
   maxHeight *= 2;
   maxDepth *= 2;

   // Get the largest axis and save as size of the bounding box.
   m_bbSize = maxWidth;
   if(m_bbSize < maxHeight) m_bbSize = maxHeight;
   if(m_bbSize < maxDepth) m_bbSize = maxDepth;
}


void COctree::CreateNode(float *vertexList, int numVerts, float *t1, float *t2,
                         float *bbCenter, float bbSize)
{
   int numTriangles = numVerts / 3;
   m_bbSize = bbSize;

   m_bbCenter[0] = bbCenter[0];
   m_bbCenter[1] = bbCenter[1];
   m_bbCenter[2] = bbCenter[2];

   // If we reached either of these limits then save the data to this
   // node.  Else create the sub nodes and keep going down the tree.
   if(numTriangles <= g_maxTriangles || g_currentLevel >= g_maxLevel)
      {
         m_isLeaf = true;
         m_numPoints = numVerts;
         m_vertices = new float[numVerts * 3];
         m_texCoord1 = new float[numVerts * 2];
         m_texCoord2 = new float[numVerts * 2];
         memcpy(m_vertices, vertexList, sizeof(float) * numVerts * 3);
         memcpy(m_texCoord1, t1, sizeof(float) * numVerts * 2);
         memcpy(m_texCoord2, t2, sizeof(float) * numVerts * 2);
      }
   else
      {
         m_isLeaf = false;

         int size1 = 0; int size2 = 0; int size3 = 0; int size4 = 0;
         int size5 = 0; int size6 = 0; int size7 = 0; int size8 = 0;
         bool *bArray1 = new bool[numTriangles];
         bool *bArray2 = new bool[numTriangles];
         bool *bArray3 = new bool[numTriangles];
         bool *bArray4 = new bool[numTriangles];
         bool *bArray5 = new bool[numTriangles];
         bool *bArray6 = new bool[numTriangles];
         bool *bArray7 = new bool[numTriangles];
         bool *bArray8 = new bool[numTriangles];
         memset(bArray1, 0, numTriangles);
         memset(bArray2, 0, numTriangles);
         memset(bArray3, 0, numTriangles);
         memset(bArray4, 0, numTriangles);
         memset(bArray5, 0, numTriangles);
         memset(bArray6, 0, numTriangles);
         memset(bArray7, 0, numTriangles);
         memset(bArray8, 0, numTriangles);

         int t = 0;

         // Loop through and determine which triangles fall within which
         // of the 8 nodes.
         for(int i = 0; i < numVerts * 3; i+=3)
            {
               // Top left front check.
               if((vertexList[i + 0] <= bbCenter[0]) && (vertexList[i + 1] >= bbCenter[1]) &&
                  (vertexList[i + 2] >= bbCenter[2])) { bArray1[t / 3] = true; }

               // Top right front check.
               if((vertexList[i + 0] >= bbCenter[0]) && (vertexList[i + 1] >= bbCenter[1]) &&
                  (vertexList[i + 2] >= bbCenter[2])) { bArray2[t / 3] = true; }

               // Top left back check.
               if((vertexList[i + 0] <= bbCenter[0]) && (vertexList[i + 1] >= bbCenter[1]) &&
                  (vertexList[i + 2] <= bbCenter[2])) { bArray3[t / 3] = true; }

               // Top right back check.
               if((vertexList[i + 0] >= bbCenter[0]) && (vertexList[i + 1] >= bbCenter[1]) &&
                  (vertexList[i + 2] <= bbCenter[2])) { bArray4[t / 3] = true; }

               // Bottom left front check.
               if((vertexList[i + 0] <= bbCenter[0]) && (vertexList[i + 1] <= bbCenter[1]) &&
                  (vertexList[i + 2] >= bbCenter[2])) { bArray5[t / 3] = true; }

               // Bottom right front check.
               if((vertexList[i + 0] >= bbCenter[0]) && (vertexList[i + 1] <= bbCenter[1]) &&
                  (vertexList[i + 2] >= bbCenter[2])) { bArray6[t / 3] = true; }

               // Bottom left back check.
               if((vertexList[i + 0] <= bbCenter[0]) && (vertexList[i + 1] <= bbCenter[1]) &&
                  (vertexList[i + 2] <= bbCenter[2])) { bArray7[t / 3] = true; }

               // Bottom right back check.
               if((vertexList[i + 0] >= bbCenter[0]) && (vertexList[i + 1] <= bbCenter[1]) &&
                  (vertexList[i + 2] <= bbCenter[2])) { bArray8[t / 3] = true; }

               t++;
            }

         // Get the number of triangles that will go in each node.
         for(i = 0; i < numTriangles; i++)
            {
               if(bArray1[i] == true) size1++;
               if(bArray2[i] == true) size2++;
               if(bArray3[i] == true) size3++;
               if(bArray4[i] == true) size4++;
               if(bArray5[i] == true) size5++;
               if(bArray6[i] == true) size6++;
               if(bArray7[i] == true) size7++;
               if(bArray8[i] == true) size8++;
            }

         // Create each sub node.
         CreateLeafNode(vertexList, numVerts, t1, t2, bArray1, size1, bbCenter, bbSize, TOP_FRONT_LEFT);
         CreateLeafNode(vertexList, numVerts, t1, t2, bArray2, size2, bbCenter, bbSize, TOP_FRONT_RIGHT);
         CreateLeafNode(vertexList, numVerts, t1, t2, bArray3, size3, bbCenter, bbSize, TOP_BACK_LEFT);
         CreateLeafNode(vertexList, numVerts, t1, t2, bArray4, size4, bbCenter, bbSize, TOP_BACK_RIGHT);
         CreateLeafNode(vertexList, numVerts, t1, t2, bArray5, size5, bbCenter, bbSize, BOTTOM_FRONT_LEFT);
         CreateLeafNode(vertexList, numVerts, t1, t2, bArray6, size6, bbCenter, bbSize, BOTTOM_FRONT_RIGHT);
         CreateLeafNode(vertexList, numVerts, t1, t2, bArray7, size7, bbCenter, bbSize, BOTTOM_BACK_LEFT);
         CreateLeafNode(vertexList, numVerts, t1, t2, bArray8, size8, bbCenter, bbSize, BOTTOM_BACK_RIGHT);

         // Free temp data.
         delete[] bArray1;
         delete[] bArray2;
         delete[] bArray3;
         delete[] bArray4;
         delete[] bArray5;
         delete[] bArray6;
         delete[] bArray7;
         delete[] bArray8;
      }
}


void COctree::CreateLeafNode(float *vList, int numVerts, float *t1, float *t2,
                             bool *bArray, int numTriangles, float *bbCenter,
                             float bbSize, int nodeId)
{
   float *vertexList = NULL, *texCoord1 = NULL, *texCoord2 = NULL;
   int vCount = 0;
   int t1Count = 0, t2Count = 0;

   if(!vList || !bArray || !bbCenter || numTriangles <= 0 ||
      numVerts <= 0 || !t1 || !t2) return;

   // Allocate temp buffers for the data.
   vertexList = new float[(numTriangles * 3) * 3];
   texCoord1 = new float[(numTriangles * 3) * 2];
   texCoord2 = new float[(numTriangles * 3) * 2];

   int t = 0;

   // Loop through and copy all vertices that fall within
   // this node in the temp buffers.
   for(int i = 0, i2 = 0; i < numVerts * 3; i+=3, i2+=2)
      {
         if(bArray[t / 3])
            {
               vertexList[vCount++] = vList[i + 0];
               vertexList[vCount++] = vList[i + 1];
               vertexList[vCount++] = vList[i + 2];
               
               texCoord1[t1Count++] = t1[i2 + 0];
               texCoord1[t1Count++] = t1[i2 + 1];

               texCoord2[t2Count++] = t2[i2 + 0];
               texCoord2[t2Count++] = t2[i2 + 1];
            }

         t++;
      }

   // Allocate this sub node.
   m_subNodes[nodeId] = new COctree();

   // Get the center of this node's bounding box.
   float *pC = GetNodeCenter(bbCenter, bbSize, nodeId);
   float center[3];
   center[0] = pC[0]; center[1] = pC[1]; center[2] = pC[2];

   g_currentLevel++;

   // Recursively keep doing down the tree.
   m_subNodes[nodeId]->CreateNode(vertexList, numTriangles * 3, texCoord1,
                                  texCoord2, center, bbSize / 2);

   g_currentLevel--;

   // Delete temp buffers.
   delete[] vertexList;
   delete[] texCoord1;
   delete[] texCoord2;
}


float *COctree::GetNodeCenter(float *bbCenter, float bbSize, int nodeId)
{
   float newCenter[3];
   newCenter[0] = newCenter[1] = newCenter[2] = 0;

   // Depending on which node we are dealing with
   // will depend on how we calculate the center.
   switch(nodeId)
      {
         case TOP_FRONT_LEFT:
            newCenter[0] = bbCenter[0] - bbSize / 4;
            newCenter[1] = bbCenter[1] + bbSize / 4;
            newCenter[2] = bbCenter[2] + bbSize / 4;
            break;

         case TOP_FRONT_RIGHT:
            newCenter[0] = bbCenter[0] + bbSize / 4;
            newCenter[1] = bbCenter[1] + bbSize / 4;
            newCenter[2] = bbCenter[2] + bbSize / 4;
            break;

         case TOP_BACK_LEFT:
            newCenter[0] = bbCenter[0] - bbSize / 4;
            newCenter[1] = bbCenter[1] + bbSize / 4;
            newCenter[2] = bbCenter[2] - bbSize / 4;
            break;

         case TOP_BACK_RIGHT:
            newCenter[0] = bbCenter[0] + bbSize / 4;
            newCenter[1] = bbCenter[1] + bbSize / 4;
            newCenter[2] = bbCenter[2] - bbSize / 4;
            break;

         case BOTTOM_FRONT_LEFT:
            newCenter[0] = bbCenter[0] - bbSize / 4;
            newCenter[1] = bbCenter[1] - bbSize / 4;
            newCenter[2] = bbCenter[2] + bbSize / 4;
            break;

         case BOTTOM_FRONT_RIGHT:
            newCenter[0] = bbCenter[0] + bbSize / 4;
            newCenter[1] = bbCenter[1] - bbSize / 4;
            newCenter[2] = bbCenter[2] + bbSize / 4;
            break;

         case BOTTOM_BACK_LEFT:
            newCenter[0] = bbCenter[0] - bbSize / 4;
            newCenter[1] = bbCenter[1] - bbSize / 4;
            newCenter[2] = bbCenter[2] - bbSize / 4;
            break;

         case BOTTOM_BACK_RIGHT:
            newCenter[0] = bbCenter[0] + bbSize / 4;
            newCenter[1] = bbCenter[1] - bbSize / 4;
            newCenter[2] = bbCenter[2] - bbSize / 4;
            break;
      }

   return &newCenter[0];
}



void COctree::Shutdown()
{
   // Release all data.

   m_numPoints = 0;
   m_isLeaf = false;

   if(m_vertices != NULL)
      {
         delete[] m_vertices;
         m_vertices = NULL;
      }

   if(m_texCoord1 != NULL)
      {
         delete[] m_texCoord1;
         m_texCoord1 = NULL;
      }

   if(m_texCoord2 != NULL)
      {
         delete[] m_texCoord2;
         m_texCoord2 = NULL;
      }

   for(int i = 0; i < 8; i++)
      {
         if(m_subNodes[i] != NULL)
            {
               m_subNodes[i]->Shutdown();
               delete m_subNodes[i];
               m_subNodes[i] = NULL;
            }
      }
}


// Copyright October 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com