/*
   Octree Terrain Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to render our terrain using a octree.
   Octrees are structures where you have a root node and 8 sub nodes.
   A sub node, in the case of octrees, are just 8 pointer objects to
   to other COctree objects.  So if you have a root node of type COctree,
   the root has pointers inside it for 8 other nodes.  These sub nodes
   also have 8 pointers in them that can be used to create another level
   in the octree and so on.  When we fill the octree we create the root
   node and create its bounding box.  If the total number of vertices
   is greater than the max number of vertices each node can hold or if
   we past the recursion limit (which in the case of a root node we would
   not have past the recursion limit yet) then we will create each of
   the 8 sub nodes and create their bounding boxes.  We keep doing
   this until we hit the recursion limit or if the number of triangles
   left is smaller than the max number of triangles we want for each node.
   When that happens we just copy the list of vertices into that node.
   When it is time to draw the octree we start with the root node.  If
   the root node does not have any vertices in it (which means that our
   mesh was too big to be stored in just 1 node and a tree was created)
   we test each sub node for vertices.  If the sub node does not have
   vertices and it does not have any sub nodes allocated, then that
   sub node has nothing in it.  Else if there are sub nodes allocated then
   we must test those to see which ones have data in them.  When we find
   nodes with data in it, we draw that data.

   It sounds more confusing than it really is.  Creating a octree would
   be pointless if we couldn't use it to eliminate geometry so you could
   always add frustum culling to the mix.  Just test the node's bounding box
   for visibility to determine if you should draw the data or not.  For this
   demo we will just concentrate on creating and rendering a octree.
*/


// Demo: Octree Terrain.


#define WIN32_LEAN_AND_MEAN// Trims down the libraries.
#define VC_LEANMEAN        // Trims even further.


#include<windows.h>
#include<gl/gl.h>
#include<gl/glu.h>
#include<gl/glext.h>
#include"CTerrain.h"
#include"CTgaLoader.h"
#include"CCamera.h"
#include"CSkyDome.h"


// Function prototypes...
bool InitializeGL();
void RenderScene();
void GetInput();
void Shutdown();


// Width and height of the screen.
#define ScreenWidth 640
#define ScreenHeight 480


// Device context.
HDC g_HDC;


// Terrain object.
CTerrain Terrain;
unsigned int terrainTexID = -1;
CTGALoader terrainDetailTex;

// Sky dome, its texture, and a camera.
CSkyDome skyDome;
CTGALoader skyTex;
CCamera Camera;

// Multitexture function.
PFNGLACTIVETEXTUREARBPROC glActiveTextureARB = NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureARB = NULL;


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         0,                               // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // This chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;                         // Rendering context.
   static HDC hDC;                           // Device context.
   int width, height;                        // The window width and height.

   switch(message)
      {
         case WM_CREATE:                     // Windows creation.
            hDC = GetDC(hwnd);               // This gets the device context for our window.
            g_HDC = hDC;                     // Assigns the global device context to this one.
            SetupPixelFormat(hDC);           // Call the pixel format function.

            hRC = wglCreateContext(hDC);     // Creates the rendering context.
            wglMakeCurrent(hDC, hRC);        // Makes the rendering context.
            return 0;
            break;


         case WM_CLOSE:                      // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);           // Deletes the rendering context.

            PostQuitMessage(0);              // Says close the program.
            return 0;
            break;


         case WM_SIZE:                       // re-size message.
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height==0)                    // we don't want it to be possible for a
               {                             // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height); // resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);     // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            // Calculate the aspect ratio of the window.
            gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 2000.0f);

            glMatrixMode(GL_MODELVIEW);      // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            return 0;
            break;


         case WM_KEYDOWN:
		      switch(wParam)
               { 
                  // Close the app if the user presses escape.
		            case VK_ESCAPE:
                     PostQuitMessage(0);
				         break;

                  default:
                     break;
               }
            break;


         default:
            break;
      }

   // What this does is pass all of the unhandled messages to DefWindowProc.
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                        // A message variable.
   WNDCLASSEX windowClass;                         // Your Window class.
   HWND hwnd;                                      // The Window handle.
   bool isFinished;                                // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   // Handle of application Icon.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   // icon when minimized.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create window...
   hwnd = CreateWindowEx(NULL,// The extended window style.
                        "UltimateGameProgrammingClass",// window Class name.
                        "Octrees Demo - By the Programming Ace.",// window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |// The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |// window style.
                        WS_CLIPSIBLINGS,// window style.
                        100, 100,// window x, y coordinate.
                        ScreenWidth, ScreenHeight,// window width and height.
                        NULL,// handle to parent window.
                        NULL,// menu.
                        hInstance,// handle to app instance.
                        NULL);   // pointer to window creation data.
   
   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.

   // If initialize fail (return false), then we don't want the program to run.
   if(!InitializeGL()) isFinished = true;

   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message then quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

   // Shutdown (free) all resources we used up.
   Shutdown();

   return (int)msg.wParam;
}


bool InitializeGL()
{
   glClearColor(0.6f, 0.6f, 1.0f, 1.0f);

   glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

   // With this we will load the list of extensions the hardware supports.
   char *extension = (char*)glGetString(GL_EXTENSIONS);
   if(strstr(extension, "GL_ARB_multitexture") == 0) return false;

   // Load the multitexture funtion and test if all went well.
   glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");
   glClientActiveTextureARB = (PFNGLCLIENTACTIVETEXTUREARBPROC)wglGetProcAddress("glClientActiveTextureARB");
   if(!glActiveTextureARB || !glClientActiveTextureARB) return false;

   // Load map.
   if(!Terrain.LoadMap("map.raw", 8)) return false;

   // Load our textures for each level.  They will be blended into one huge texture
   // based on the height data of the terrain.
   if(!Terrain.LoadTile(LOWEST_TILE, "lowest.tga")) return false;
	if(!Terrain.LoadTile(LOW_TILE, "low.tga")) return false;
	if(!Terrain.LoadTile(HIGH_TILE, "high.tga")) return false;
	if(!Terrain.LoadTile(HIGHEST_TILE, "highest.tga")) return false;

   unsigned char *image = NULL;
   int imgSize = 0;

   // Generate the terrain texture.
   image = Terrain.GenerateTextureMap(256);
   if(!image) return false;

   // Load both textures.
   glGenTextures(1, &terrainTexID);
	glBindTexture(GL_TEXTURE_2D, terrainTexID);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);						
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 256, 256,
	             0, GL_RGB, GL_UNSIGNED_BYTE, image);

   delete[] image;

   if(!terrainDetailTex.LoadTGA("detail.tga")) return false;
   glGenTextures(1, &terrainDetailTex.ID);
	glBindTexture(GL_TEXTURE_2D, terrainDetailTex.ID);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);						
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, terrainDetailTex.imageWidth,
	             terrainDetailTex.imageHeight, 0, GL_RGB, GL_UNSIGNED_BYTE,
	             terrainDetailTex.image);

   // Set the camera's default position.
   Camera.SetCamera(50.0f, 50.0f, 10.0f, 50.0f, 50.0f, 11.0f, 0, 1, 0);

   // Initialize the sky dome and load its texture.
   if(!skyDome.Initialize(15.0f, 15.0f, 1024.0f)) return false;

   if(!skyTex.LoadTGA("clouds.tga")) return false;
   glGenTextures(1, &skyTex.ID);
	glBindTexture(GL_TEXTURE_2D, skyTex.ID);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);						
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, skyTex.imageWidth,
	             skyTex.imageHeight, 0, GL_RGB, GL_UNSIGNED_BYTE,
	             skyTex.image);

   return true;
}


void DrawBoundingBox(float *min, float *max)
{
	glColor3f(1.0, 0.0, 0.0f);
   glBindTexture(GL_TEXTURE_2D, NULL);

   // Draw the box so we can see it.
	glBegin(GL_LINE_LOOP);
	   glVertex3f(min[0], min[1], min[2]); glVertex3f(max[0], min[1], min[2]);
	   glVertex3f(max[0], min[1], max[2]); glVertex3f(min[0], min[1], max[2]);
	glEnd();

	glBegin(GL_LINE_LOOP);
	   glVertex3f(min[0], max[1], min[2]); glVertex3f(max[0], max[1], min[2]);
	   glVertex3f(max[0], max[1], max[2]); glVertex3f(min[0], max[1], max[2]);
	glEnd();

	glBegin(GL_LINES);
	   glVertex3f(min[0], min[1], min[2]); glVertex3f(min[0], max[1], min[2]);
	   glVertex3f(max[0], min[1], min[2]); glVertex3f(max[0], max[1], min[2]);
   glEnd();

   glBegin(GL_LINES);
	   glVertex3f(max[0], min[1], max[2]); glVertex3f(max[0], max[1], max[2]);
	   glVertex3f(min[0], min[1], max[2]); glVertex3f(min[0], max[1], max[2]);
	glEnd();

   glColor3f(1.0, 1.0, 1.0f);
}


void RenderOctreeNode(COctree* pNode)
{
   if(pNode == NULL) return;

   if(pNode->isLeaf())
      {
         if(!pNode->isVertices()) return;

         int numTris = pNode->GetNumTriangles();

         // Get the min and max bounding box for this node.
         float bbSize = pNode->GetBoundingBoxSize();

         float bbMin[3];
         pNode->GetBoundingBoxCenter(bbMin[0], bbMin[1], bbMin[2]);
         bbMin[0] -= bbSize / 2.0f;
         bbMin[1] -= bbSize / 2.0f;
         bbMin[2] -= bbSize / 2.0f;

         float bbMax[3];
         pNode->GetBoundingBoxCenter(bbMax[0], bbMax[1], bbMax[2]);
         bbMax[0] += bbSize / 2.0f;
         bbMax[1] += bbSize / 2.0f;
         bbMax[2] += bbSize / 2.0f;

         // Draw the bounding box for debug purposes.
         DrawBoundingBox(bbMin, bbMax);

         // Get the data so we can render it out.
         float *pVerts = pNode->GetVertices();
         float *pTC1 = pNode->GetTexCoord1();
         float *pTC2 = pNode->GetTexCoord2();

         // Apply the textures.
         glActiveTextureARB(GL_TEXTURE0_ARB);
         glEnable(GL_TEXTURE_2D);
         glBindTexture(GL_TEXTURE_2D, terrainTexID);

         glActiveTextureARB(GL_TEXTURE1_ARB);
         glEnable(GL_TEXTURE_2D);
         glBindTexture(GL_TEXTURE_2D, terrainDetailTex.ID);

         glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_ARB);
         glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE_ARB, 2);

         glActiveTextureARB(GL_TEXTURE0_ARB);

         // Set pointers.
         glEnableClientState(GL_VERTEX_ARRAY);
         glEnableClientState(GL_TEXTURE_COORD_ARRAY);

         glVertexPointer(3, GL_FLOAT, 0, pVerts);
         glTexCoordPointer(2, GL_FLOAT, 0, pTC1);
         
         glClientActiveTextureARB(GL_TEXTURE1_ARB);
         glEnableClientState(GL_TEXTURE_COORD_ARRAY);
         glTexCoordPointer(2, GL_FLOAT, 0, pTC2);
         glClientActiveTextureARB(GL_TEXTURE0_ARB);

         // Draw the entire node's data.
         glDrawArrays(GL_TRIANGLES, 0, numTris * 3);

         // Disable all the client states we enabled.
         glDisableClientState(GL_VERTEX_ARRAY);
         glDisableClientState(GL_TEXTURE_COORD_ARRAY);
         glClientActiveTextureARB(GL_TEXTURE1_ARB);
         glDisableClientState(GL_TEXTURE_COORD_ARRAY);
         glClientActiveTextureARB(GL_TEXTURE0_ARB);
      }
   else
      {
         // Else if this node is not a leaf then
         // we continue down the tree.
         COctree **pSubNodes = pNode->GetSubNodes();
         RenderOctreeNode(pSubNodes[TOP_FRONT_LEFT]);
         RenderOctreeNode(pSubNodes[TOP_FRONT_RIGHT]);
         RenderOctreeNode(pSubNodes[TOP_BACK_LEFT]);
         RenderOctreeNode(pSubNodes[TOP_BACK_RIGHT]);

         RenderOctreeNode(pSubNodes[BOTTOM_FRONT_LEFT]);
         RenderOctreeNode(pSubNodes[BOTTOM_FRONT_RIGHT]);
         RenderOctreeNode(pSubNodes[BOTTOM_BACK_LEFT]);
         RenderOctreeNode(pSubNodes[BOTTOM_BACK_RIGHT]);
      }
}


void RenderScene()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // Clears the screen.
   glLoadIdentity();                   //Reset modelview matrix for new frame.

   GetInput();

	// Set the camera.
   gluLookAt(Camera.xPos, Camera.yPos, Camera.zPos,
             Camera.xView, Camera.yView, Camera.zView,
             Camera.xUp, Camera.yUp, Camera.zUp);


   // Render the sky dome.
   glPushMatrix();
      
      glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

      // Position the dome to always be around the cameras position.
		glTranslatef(Camera.xPos, Camera.yPos - 400.0f, Camera.zPos);
		
		// Set up the sky dome texture.
	   glBindTexture(GL_TEXTURE_2D, skyTex.ID);
	   glEnable(GL_TEXTURE_2D);
		
		// Draw a huge quad to cover bottom of dome.
		glBegin(GL_QUADS);
         glTexCoord2f(0, 0); glVertex3f(1024, 0, 1024);
         glTexCoord2f(1, 0); glVertex3f(-1024, 0, 1024);
         glTexCoord2f(1, 1); glVertex3f(-1024, 0, -1024);
         glTexCoord2f(0, 1); glVertex3f(1024, 0, -1024);
		glEnd();

		// Move the dome so the half a sphere is not facing the side but on top of everything.
		glRotatef(270, 1.0f, 0.0f, 0.0f);

		// Use arrays to send data to the hardware.
		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, 0, skyDome.vertices);

		// Send texture coords to hardware.
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);	
		glTexCoordPointer(2, GL_FLOAT, 0, skyDome.texCoords);

		// Draw the whole thing out.
		glDrawArrays(GL_TRIANGLE_STRIP, 0, skyDome.numPoints);

		// Disable the arrays since we dont need them anymore.
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		
   glPopMatrix();

   // Render the entire octree staring with the root node.
   RenderOctreeNode(Terrain.m_rootNode);

	SwapBuffers(g_HDC);
}


void GetInput()
{
   // Checking for input.  We move the camera based on the button(s) that were pressed.
   if(GetKeyState(VK_UP) & 0x80) Camera.MoveCamera(UP);
   if(GetKeyState(VK_DOWN) & 0x80) Camera.MoveCamera(DOWN);

   if(GetKeyState(VK_LEFT) & 0x80) Camera.StrafeCam(STRAFE_LEFT);
   if(GetKeyState(VK_RIGHT) & 0x80) Camera.StrafeCam(STRAFE_RIGHT);

   // Here we get the position of the mouse, rotate the camera based on its movements,
   // then set the mouse position to the middle of the screen.
   POINT mousePos;
	GetCursorPos(&mousePos);

   Camera.RotateByMouse(mousePos.x, mousePos.y, (ScreenWidth >> 1), (ScreenHeight >> 1));
   Camera.RotateByMouse(mousePos.x, mousePos.y, (ScreenWidth >> 1), (ScreenHeight >> 1));

	SetCursorPos((ScreenWidth >> 1), (ScreenHeight >> 1));
}


void Shutdown()
{
   // Release all resources our terrain uses.
   skyDome.ShutDown();
   skyTex.FreeImage();
   Terrain.ShutDown();
   terrainDetailTex.FreeImage();
   glDeleteTextures(1, &terrainTexID);
}


// Recap:

// Rendering a mesh using octrees is a lot easier than I am sure many
// people thought.  All we are doing is creating a bounding box around
// the entire terrain.  Then we break that box up into 8 equally sized
// cubes within the box.  We then determine which vertices fall within
// with cubes.  Depending of if we reached the max recursion level or
// if we reached the smallest number of triangles allowed in each cube
// will depend on if we break each cube into another, smaller set of 8
// or if we are done creating the tree.  Technically you can put any mesh
// in a octree, not just terrains.  It might not make sense for character
// models but you can if you want.


// Copyright October 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com